@@ -87,86 +87,6 @@ class Utilities(script_utilities.Utilities):
 
         return index
 
-    def displayedText(self, obj):
-        """Returns the text being displayed for an object.
-
-        Arguments:
-        - obj: the object
-
-        Returns the text being displayed for an object or None if there isn't
-        any text being shown.  Overridden in this script because we have lots
-        of whitespace we need to remove.
-        """
-
-        displayedText = script_utilities.Utilities.displayedText(self, obj)
-        if displayedText \
-           and not (obj.getState().contains(pyatspi.STATE_EDITABLE) \
-                    or obj.getRole() in [pyatspi.ROLE_ENTRY,
-                                         pyatspi.ROLE_PASSWORD_TEXT]):
-            displayedText = displayedText.strip()
-            # Some ARIA widgets (e.g. the list items in the chat box
-            # in gmail) implement the accessible text interface but
-            # only contain whitespace.
-            #
-            if not displayedText \
-               and obj.getState().contains(pyatspi.STATE_FOCUSED):
-                label = self.displayedLabel(obj)
-                if not label:
-                    displayedText = obj.name
-
-        return displayedText
-
-    def displayedLabel(self, obj):
-        """If there is an object labelling the given object, return the
-        text being displayed for the object labelling this object.
-        Otherwise, return None.  Overridden here to handle instances
-        of bogus labels and form fields where a lack of labels necessitates
-        our attempt to guess the text that is functioning as a label.
-
-        Argument:
-        - obj: the object in question
-
-        Returns the string of the object labelling this object, or None
-        if there is nothing of interest here.
-        """
-
-        string = None
-        labels = self.labelsForObject(obj)
-        for label in labels:
-            # Check to see if the official labels are valid.
-            #
-            bogus = False
-            if self._script.inDocumentContent() \
-               and obj.getRole() in [pyatspi.ROLE_COMBO_BOX,
-                                     pyatspi.ROLE_LIST_BOX,
-                                     pyatspi.ROLE_LIST]:
-                # Bogus case #1:
-                # <label></label> surrounding the entire combo box/list which
-                # makes the entire combo box's/list's contents serve as the
-                # label. We can identify this case because the child of the
-                # label is the combo box/list. See bug #428114, #441476.
-                #
-                if label.childCount:
-                    bogus = (label[0].getRole() == obj.getRole())
-
-            if not bogus:
-                # Bogus case #2:
-                # <label></label> surrounds not just the text serving as the
-                # label, but whitespace characters as well (e.g. the text
-                # serving as the label is on its own line within the HTML).
-                # Because of the Mozilla whitespace bug, these characters
-                # will become part of the label which will cause the label
-                # and name to no longer match and Orca to seemingly repeat
-                # the label.  Therefore, strip out surrounding whitespace.
-                # See bug #441610 and
-                # https://bugzilla.mozilla.org/show_bug.cgi?id=348901
-                #
-                expandedLabel = self.expandEOCs(label)
-                if expandedLabel:
-                    string = self.appendString(string, expandedLabel.strip())
-
-        return string
-
     def documentFrame(self):
         """Returns the document frame that holds the content being shown."""
 
@@ -239,7 +159,8 @@ class Utilities(script_utilities.Utilities):
         if not obj:
             obj = orca_state.locusOfFocus
 
-        if obj and obj.parent.getRole() == pyatspi.ROLE_AUTOCOMPLETE:
+        if obj and obj.parent \
+           and obj.parent.getRole() == pyatspi.ROLE_AUTOCOMPLETE:
             return False
 
         return script_utilities.Utilities.inFindToolbar(obj)
@@ -261,6 +182,23 @@ class Utilities(script_utilities.Utilities):
 
         return False
 
+    def isLink(self, obj):
+        """Returns True if we should treat this object as a link."""
+
+        if not obj:
+            return False
+
+        role = obj.getRole()
+        if role == pyatspi.ROLE_LINK:
+            return True
+
+        if role == pyatspi.ROLE_TEXT \
+           and obj.parent.getRole() == pyatspi.ROLE_LINK \
+           and obj.name and obj.name == obj.parent.name:
+            return True
+
+        return False
+
     def isPasswordText(self, obj):
         """Returns True if we should treat this object as password text."""
 
@@ -570,6 +508,7 @@ class Utilities(script_utilities.Utilities):
         if not obj or not contents:
             return -1
 
+        offset = max(0, offset)
         matches = [x for x in contents if self.isSameObject(x[0], obj)]
         match = [x for x in matches if x[1] <= offset < x[2]]
         if match and match[0] and match[0] in contents:
@@ -583,7 +522,8 @@ class Utilities(script_utilities.Utilities):
                  pyatspi.ROLE_MENU_ITEM,
                  pyatspi.ROLE_RADIO_MENU_ITEM,
                  pyatspi.ROLE_RADIO_BUTTON,
-                 pyatspi.ROLE_PUSH_BUTTON]
+                 pyatspi.ROLE_PUSH_BUTTON,
+                 pyatspi.ROLE_TOGGLE_BUTTON]
 
         role = obj.getRole()
         if role in roles:
@@ -687,27 +627,43 @@ class Utilities(script_utilities.Utilities):
             debug.println(debug.LEVEL_INFO, msg)
             return sadString, sadStart, sadEnd
 
-        if not boundary == pyatspi.TEXT_BOUNDARY_LINE_START:
-            return text.getText(start, end), start, end
-
         return text.getText(start, end), start, end
 
-    def _getWordContentsForObj(self, obj, offset):
+    def _getContentsForObj(self, obj, offset, boundary):
         if not obj:
             return []
 
-        boundary = pyatspi.TEXT_BOUNDARY_WORD_START
         string, start, end = self._getTextAtOffset(obj, offset, boundary)
-        if string and string.find(self.EMBEDDED_OBJECT_CHARACTER) == -1:
+        if not string:
             return [[obj, start, end, string]]
 
-        return self.getObjectsFromEOCs(obj, start, boundary)
+        stringOffset = offset - start
+        try:
+            char = string[stringOffset]
+        except:
+            pass
+        else:
+            if char == self.EMBEDDED_OBJECT_CHARACTER:
+                childIndex = self._script.getChildIndex(obj, offset)
+                child = obj[childIndex]
+                return [[child, 0, 1, ""]]
+
+        ranges = [m.span() for m in re.finditer("[^\ufffc]+", string)]
+        strings = list(filter(lambda x: x[0] <= stringOffset <= x[1], ranges))
+        if len(strings) == 1:
+            rangeStart, rangeEnd = strings[0]
+            start += rangeStart
+            string = string[rangeStart:rangeEnd]
+            end = start + len(string)
+
+        return [[obj, start, end, string]]
 
     def getWordContentsAtOffset(self, obj, offset):
         if not obj:
             return []
 
-        objects = self._getWordContentsForObj(obj, offset)
+        boundary = pyatspi.TEXT_BOUNDARY_WORD_START
+        objects = self._getContentsForObj(obj, offset, boundary)
         extents = self.getExtents(obj, offset, offset + 1)
 
         def _include(x):
@@ -729,7 +685,7 @@ class Utilities(script_utilities.Utilities):
             if not text or text.getText(pOffset, pOffset + 1).isspace():
                 break
 
-            onLeft = self._getWordContentsForObj(prevObj, pOffset)
+            onLeft = self._getContentsForObj(prevObj, pOffset, boundary)
             onLeft = list(filter(_include, onLeft))
             if not onLeft:
                 break
@@ -742,7 +698,7 @@ class Utilities(script_utilities.Utilities):
         lastObj, lastStart, lastEnd, lastString = objects[-1]
         while lastObj and lastString and not lastString[-1].isspace():
             nextObj, nOffset = self._script.findNextCaretInOrder(lastObj, lastEnd - 1)
-            onRight = self._getWordContentsForObj(nextObj, nOffset)
+            onRight = self._getContentsForObj(nextObj, nOffset, boundary)
             onRight = list(filter(_include, onRight))
             if not onRight:
                 break
@@ -757,23 +713,17 @@ class Utilities(script_utilities.Utilities):
 
         return objects
 
-    def _getLineContentsForObj(self, obj, offset):
-        if not obj:
-            return []
+    def _contentIsSubsetOf(self, contentA, contentB):
+        objA, startA, endA, stringA = contentA
+        objB, startB, endB, stringB = contentB
+        if objA == objB:
+            setA = set(range(startA, endA))
+            setB = set(range(startB, endB))
+            return setA.issubset(setB)
 
-        boundary = pyatspi.TEXT_BOUNDARY_LINE_START
-        string, start, end = self._getTextAtOffset(obj, offset, boundary)
+        return False
 
-        while string.startswith(self.EMBEDDED_OBJECT_CHARACTER):
-            string = string[1:]
-            start += 1
-
-        if string and string.find(self.EMBEDDED_OBJECT_CHARACTER) == -1:
-            return [[obj, start, end, string]]
-
-        return self.getObjectsFromEOCs(obj, start, boundary)
-
-    def getLineContentsAtOffset(self, obj, offset):
+    def getLineContentsAtOffset(self, obj, offset, layoutMode=True):
         if not obj:
             return []
 
@@ -791,26 +741,45 @@ class Utilities(script_utilities.Utilities):
             xExtents = self.getExtents(xObj, xStart, xStart + 1)
             return self.extentsAreOnSameLine(extents, xExtents)
 
-        objects = self._getLineContentsForObj(obj, offset)
+        boundary = pyatspi.TEXT_BOUNDARY_LINE_START
+        objects = self._getContentsForObj(obj, offset, boundary)
+
+        firstObj, firstStart, firstEnd, firstString = objects[0]
+        lastObj, lastStart, lastEnd, lastString = objects[-1]
+        prevObj, pOffset = self._script.findPreviousCaretInOrder(firstObj, firstStart)
+        nextObj, nOffset = self._script.findNextCaretInOrder(lastObj, lastEnd - 1)
+        if not layoutMode:
+            if firstString and not re.search("\w", firstString) \
+               and (re.match("[^\w\s]", firstString[0]) or not firstString.strip()):
+                onLeft = self._getContentsForObj(prevObj, pOffset, boundary)
+                onLeft = list(filter(_include, onLeft))
+                objects[0:0] = onLeft
+
+            char = self._script.getCharacterAtOffset(nextObj, nOffset)
+            if re.match("[^\w\s]", char):
+                objects.append([nextObj, nOffset, nOffset + 1, char])
+
+            return objects
 
         # Check for things on the same line to the left of this object.
-        firstObj, firstStart = objects[0][0], objects[0][1]
-        prevObj, pOffset = self._script.findPreviousCaretInOrder(firstObj, firstStart)
         while prevObj:
-            onLeft = self._getLineContentsForObj(prevObj, pOffset)
+            onLeft = self._getContentsForObj(prevObj, pOffset, boundary)
             onLeft = list(filter(_include, onLeft))
             if not onLeft:
                 break
 
+            if self._contentIsSubsetOf(objects[0], onLeft[-1]):
+                objects.pop(0)
+
             objects[0:0] = onLeft
             firstObj, firstStart = objects[0][0], objects[0][1]
             prevObj, pOffset = self._script.findPreviousCaretInOrder(firstObj, firstStart)
+            if self._script.getCharacterAtOffset(prevObj, pOffset) in [" ", "\xa0"]:
+                prevObj, pOffset = self._script.findPreviousCaretInOrder(prevObj, pOffset)
 
         # Check for things on the same line to the right of this object.
-        lastObj, lastEnd = objects[-1][0], objects[-1][2]
-        nextObj, nOffset = self._script.findNextCaretInOrder(lastObj, lastEnd - 1)
         while nextObj:
-            onRight = self._getLineContentsForObj(nextObj, nOffset)
+            onRight = self._getContentsForObj(nextObj, nOffset, boundary)
             onRight = list(filter(_include, onRight))
             if not onRight:
                 break
@@ -818,9 +787,73 @@ class Utilities(script_utilities.Utilities):
             objects.extend(onRight)
             lastObj, lastEnd = objects[-1][0], objects[-1][2]
             nextObj, nOffset = self._script.findNextCaretInOrder(lastObj, lastEnd - 1)
+            if self._script.getCharacterAtOffset(nextObj, nOffset) in [" ", "\xa0"]:
+                nextObj, nOffset = self._script.findNextCaretInOrder(nextObj, nOffset)
 
         return objects
 
+    def isTextBlockElement(self, obj):
+        if not self._script.inDocumentContent(obj):
+            return False
+
+        textBlockElements = [pyatspi.ROLE_COLUMN_HEADER,
+                             pyatspi.ROLE_DOCUMENT_FRAME,
+                             pyatspi.ROLE_FORM,
+                             pyatspi.ROLE_HEADING,
+                             pyatspi.ROLE_LIST,
+                             pyatspi.ROLE_LIST_ITEM,
+                             pyatspi.ROLE_PANEL,
+                             pyatspi.ROLE_PARAGRAPH,
+                             pyatspi.ROLE_ROW_HEADER,
+                             pyatspi.ROLE_SECTION,
+                             pyatspi.ROLE_TEXT,
+                             pyatspi.ROLE_TABLE_CELL]
+
+        role = obj.getRole()
+        if not role in textBlockElements:
+            return False
+
+        state = obj.getState()
+        if state.contains(pyatspi.STATE_EDITABLE):
+            return False
+
+        if not state.contains(pyatspi.STATE_FOCUSABLE) \
+           or role == pyatspi.ROLE_DOCUMENT_FRAME:
+            return True
+
+        return False
+
+    def filterContentsForPresentation(self, contents, inferLabels=False):
+        def _include(x):
+            obj, start, end, string = x
+            if not obj:
+                return False
+
+            if (self.isTextBlockElement(obj) and not string.strip()) \
+               or self.isLabellingContents(x, contents) \
+               or self.isOffScreenLabel(obj, start):
+                return False
+
+            if inferLabels and self.isInferredLabelForContents(x, contents):
+                return False
+
+            return True
+
+        return list(filter(_include, contents))
+
+    def needsSeparator(self, lastChar, nextChar):
+        if lastChar.isspace() or nextChar.isspace():
+            return False
+
+        openingPunctuation = ["(", "[", "{", "<"]
+        closingPunctuation = [".", "?", "!", ":", ",", ";", ")", "]", "}", ">"]
+        if lastChar in closingPunctuation or nextChar in openingPunctuation:
+            return True
+        if lastChar in openingPunctuation or nextChar in closingPunctuation:
+            return False
+
+        return lastChar.isalnum()
+
     def getObjectsFromEOCs(self, obj, offset=None, boundary=None):
         """Expands the current object replacing EMBEDDED_OBJECT_CHARACTERS
         with [obj, startOffset, endOffset, string] tuples.
@@ -920,6 +953,71 @@ class Utilities(script_utilities.Utilities):
 
         return objects
 
+    def isOffScreenLabel(self, obj, offset=0):
+        if not obj:
+            return False
+
+        isLabelFor = lambda x: x.getRelationType() == pyatspi.RELATION_LABEL_FOR
+        relations = list(filter(isLabelFor, obj.getRelationSet()))
+        if relations:
+            offset = max(offset, 0)
+            x, y, width, height = self.getExtents(obj, offset, offset + 1)
+            if x < 0 or y < 0:
+                return True
+
+        return False
+
+    def shouldInferLabelFor(self, obj):
+        if self.displayedLabel(obj) or obj.name:
+            return False
+
+        roles =  [pyatspi.ROLE_CHECK_BOX,
+                  pyatspi.ROLE_COMBO_BOX,
+                  pyatspi.ROLE_ENTRY,
+                  pyatspi.ROLE_LIST_BOX,
+                  pyatspi.ROLE_PASSWORD_TEXT,
+                  pyatspi.ROLE_RADIO_BUTTON]
+        if not obj.getRole() in roles:
+            return False
+
+        if not self._script.inDocumentContent():
+            return False
+
+        return True
+
+    def isInferredLabelForContents(self, content, contents):
+        obj, start, end, string = content
+        objs = list(filter(self.shouldInferLabelFor, [x[0] for x in contents]))
+        if not objs:
+            return None
+
+        for o in objs:
+            label, sources = self._script.labelInference.infer(o, False)
+            if obj in sources and label.strip() == string.strip():
+                return o
+
+        return None
+
+    def isLabellingContents(self, content, contents):
+        obj, start, end, string = content
+        if obj.getRole() != pyatspi.ROLE_LABEL:
+            return None
+
+        relationSet = obj.getRelationSet()
+        if not relationSet:
+            return None
+
+        for relation in relationSet:
+            if relation.getRelationType() \
+                == pyatspi.RELATION_LABEL_FOR:
+                for i in range(0, relation.getNTargets()):
+                    target = relation.getTarget(i)
+                    for content in contents:
+                        if content[0] == target:
+                            return target
+
+        return None
+
     def isClickableElement(self, obj):
         if not self._script.inDocumentContent(obj):
             return False
@@ -957,3 +1055,37 @@ class Utilities(script_utilities.Utilities):
                 return True
 
         return False
+
+    def isZombie(self, obj):
+        try:
+            index = obj.getIndexInParent()
+            state = obj.getState()
+        except:
+            debug.println(debug.LEVEL_INFO, "ZOMBIE: %s is null or dead" % obj)
+            return True
+
+        if obj.getIndexInParent() == -1:
+            debug.println(debug.LEVEL_INFO, "ZOMBIE: %s's index is -1" % obj)
+            return True
+        if state.contains(pyatspi.STATE_DEFUNCT):
+            debug.println(debug.LEVEL_INFO, "ZOMBIE: %s is defunct" % obj)
+            return True
+        if state.contains(pyatspi.STATE_INVALID):
+            debug.println(debug.LEVEL_INFO, "ZOMBIE: %s is invalid" % obj)
+            return True
+
+        return False
+
+    def findReplicant(self, root, obj):
+        if not (root and obj):
+            return None
+
+        isSame = lambda x: x and self.isSameObject(x, obj, comparePaths=True)
+        if isSame(root):
+            replicant = root
+        else:
+            replicant = pyatspi.utils.findDescendant(root, isSame)
+
+        msg = "HACK: Returning %s as replicant for Zombie %s" % (replicant, obj)
+        debug.println(debug.LEVEL_INFO, msg)
+        return replicant
